import time

# Since we want API users to be able to provide timestamps,
# this needs to be in the API.

try:
    time_ns = time.time_ns
# Python versions < 3.7
except AttributeError:

    def time_ns() -> int:
        return int(time.time() * 1e9)



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The OpenTelemetry loader module is mainly used internally to load the
implementation for global objects like
:func:`opentelemetry.trace.tracer_source`.

.. _loader-factory:

An instance of a global object of type ``T`` is always created with a factory
function with the following signature::

    def my_factory_for_t(api_type: typing.Type[T]) -> typing.Optional[T]:
        # ...

That function is called with e.g., the type of the global object it should
create as an argument (e.g. the type object
:class:`opentelemetry.trace.TracerSource`) and should return an instance of that type
(such that ``instanceof(my_factory_for_t(T), T)`` is true). Alternatively, it
may return ``None`` to indicate that the no-op default should be used.

When loading an implementation, the following algorithm is used to find a
factory function or other means to create the global object:

    1. If the environment variable
       :samp:`OPENTELEMETRY_PYTHON_IMPLEMENTATION_{getter-name}` (e.g.,
       ``OPENTELEMETRY_PYTHON_IMPLEMENTATION_TRACERSOURCE``) is set to an
       nonempty value, an attempt is made to import a module with that name and
       use a factory function named ``get_opentelemetry_implementation`` in it.
    2. Otherwise, the same is tried with the environment variable
       ``OPENTELEMETRY_PYTHON_IMPLEMENTATION_DEFAULT``.
    3. Otherwise, if a :samp:`set_preferred_{<type>}_implementation` was
       called (e.g.
       :func:`opentelemetry.trace.set_preferred_tracer_source_implementation`),
       the callback set there is used (that is, the environment variables
       override the callback set in code).
    4. Otherwise, if :func:`set_preferred_default_implementation` was called,
       the callback set there is used.
    5. Otherwise, an attempt is made to import and use the OpenTelemetry SDK.
    6. Otherwise the default implementation that ships with the API
       distribution (a fast no-op implementation) is used.

If any of the above steps fails (e.g., a module is loaded but does not define
the required function or a module name is set but the module fails to load),
the search immediatelly skips to the last step.

Note that the first two steps (those that query environment variables) are
skipped if :data:`sys.flags` has ``ignore_environment`` set (which usually
means that the Python interpreter was invoked with the ``-E`` or ``-I`` flag).
"""

import importlib
import os
import sys
from typing import Callable, Optional, Type, TypeVar

_T = TypeVar("_T")

# "Untrusted" because this is usually user-provided and we don't trust the user
# to really return a _T: by using object, mypy forces us to check/cast
# explicitly.
_UntrustedImplFactory = Callable[[Type[_T]], Optional[object]]


# This would be the normal ImplementationFactory which would be used to
# annotate setters, were it not for https://github.com/python/mypy/issues/7092
# Once that bug is resolved, setters should use this instead of duplicating the
# code.
# ImplementationFactory = Callable[[Type[_T]], Optional[_T]]

_DEFAULT_FACTORY = None  # type: Optional[_UntrustedImplFactory[object]]


def _try_load_impl_from_modname(
    implementation_modname: str, api_type: Type[_T]
) -> Optional[_T]:
    try:
        implementation_mod = importlib.import_module(implementation_modname)
    except (ImportError, SyntaxError):
        # TODO Log/warn
        return None

    return _try_load_impl_from_mod(implementation_mod, api_type)


def _try_load_impl_from_mod(
    implementation_mod: object, api_type: Type[_T]
) -> Optional[_T]:

    try:
        # Note: We use such a long name to avoid calling a function that is not
        # intended for this API.

        implementation_fn = getattr(
            implementation_mod, "get_opentelemetry_implementation"
        )  # type: _UntrustedImplFactory[_T]
    except AttributeError:
        # TODO Log/warn
        return None

    return _try_load_impl_from_callback(implementation_fn, api_type)


def _try_load_impl_from_callback(
    implementation_fn: _UntrustedImplFactory[_T], api_type: Type[_T]
) -> Optional[_T]:
    result = implementation_fn(api_type)
    if result is None:
        return None
    if not isinstance(result, api_type):
        # TODO Warn if wrong type is returned
        return None

    # TODO: Warn if implementation_fn returns api_type(): It should return None
    # to indicate using the default.

    return result


def _try_load_configured_impl(
    api_type: Type[_T], factory: Optional[_UntrustedImplFactory[_T]]
) -> Optional[_T]:
    """Attempts to find any specially configured implementation. If none is
    configured, or a load error occurs, returns `None`
    """
    implementation_modname = None
    if not sys.flags.ignore_environment:
        implementation_modname = os.getenv(
            "OPENTELEMETRY_PYTHON_IMPLEMENTATION_" + api_type.__name__.upper()
        )
        if implementation_modname:
            return _try_load_impl_from_modname(
                implementation_modname, api_type
            )
        implementation_modname = os.getenv(
            "OPENTELEMETRY_PYTHON_IMPLEMENTATION_DEFAULT"
        )
        if implementation_modname:
            return _try_load_impl_from_modname(
                implementation_modname, api_type
            )
    if factory is not None:
        return _try_load_impl_from_callback(factory, api_type)
    if _DEFAULT_FACTORY is not None:
        return _try_load_impl_from_callback(_DEFAULT_FACTORY, api_type)
    return None


# Public to other opentelemetry-api modules
def _load_impl(
    api_type: Type[_T], factory: Optional[Callable[[Type[_T]], Optional[_T]]]
) -> _T:
    """Tries to load a configured implementation, if unsuccessful, returns a
    fast no-op implemenation that is always available.
    """

    result = _try_load_configured_impl(api_type, factory)
    if result is None:
        return api_type()
    return result


def set_preferred_default_implementation(
    implementation_factory: _UntrustedImplFactory[_T],
) -> None:
    """Sets a factory function that may be called for any implementation
    object. See the :ref:`module docs <loader-factory>` for more details."""
    global _DEFAULT_FACTORY  # pylint:disable=global-statement
    _DEFAULT_FACTORY = implementation_factory



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import typing

from opentelemetry.trace import SpanContext


class BinaryFormat(abc.ABC):
    """API for serialization of span context into binary formats.

    This class provides an interface that enables converting span contexts
    to and from a binary format.
    """

    @staticmethod
    @abc.abstractmethod
    def to_bytes(context: SpanContext) -> bytes:
        """Creates a byte representation of a SpanContext.

        to_bytes should read values from a SpanContext and return a data
        format to represent it, in bytes.

        Args:
            context: the SpanContext to serialize

        Returns:
            A bytes representation of the SpanContext.

        """

    @staticmethod
    @abc.abstractmethod
    def from_bytes(byte_representation: bytes) -> typing.Optional[SpanContext]:
        """Return a SpanContext that was represented by bytes.

        from_bytes should return back a SpanContext that was constructed from
        the data serialized in the byte_representation passed. If it is not
        possible to read in a proper SpanContext, return None.

        Args:
            byte_representation: the bytes to deserialize

        Returns:
            A bytes representation of the SpanContext if it is valid.
            Otherwise return None.

        """



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import typing


class Resource(abc.ABC):
    """The interface that resources must implement."""

    @staticmethod
    @abc.abstractmethod
    def create(labels: typing.Dict[str, str]) -> "Resource":
        """Create a new resource.

        Args:
            labels: the labels that define the resource

        Returns:
            The resource with the labels in question

        """

    @property
    @abc.abstractmethod
    def labels(self) -> typing.Dict[str, str]:
        """Return the label dictionary associated with this resource.

        Returns:
            A dictionary with the labels of the resource

        """

    @abc.abstractmethod
    def merge(self, other: typing.Optional["Resource"]) -> "Resource":
        """Return a resource with the union of labels for both resources.

        Labels that exist in the main Resource take precedence unless the
        label value is the empty string.

        Args:
            other: The resource to merge in

        """



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import typing
from os import environ
from sys import version_info

from pkg_resources import iter_entry_points

from opentelemetry.context.context import Context, RuntimeContext

logger = logging.getLogger(__name__)
_RUNTIME_CONTEXT = None  # type: typing.Optional[RuntimeContext]


def get_value(key: str, context: typing.Optional[Context] = None) -> "object":
    """To access the local state of a concern, the RuntimeContext API
    provides a function which takes a context and a key as input,
    and returns a value.

    Args:
        key: The key of the value to retrieve.
        context: The context from which to retrieve the value, if None, the current context is used.
    """
    return context.get(key) if context is not None else get_current().get(key)


def set_value(
    key: str, value: "object", context: typing.Optional[Context] = None
) -> Context:
    """To record the local state of a cross-cutting concern, the
    RuntimeContext API provides a function which takes a context, a
    key, and a value as input, and returns an updated context
    which contains the new value.

    Args:
        key: The key of the entry to set
        value: The value of the entry to set
        context: The context to copy, if None, the current context is used
    """
    if context is None:
        context = get_current()
    new_values = context.copy()
    new_values[key] = value
    return Context(new_values)


def remove_value(
    key: str, context: typing.Optional[Context] = None
) -> Context:
    """To remove a value, this method returns a new context with the key
    cleared. Note that the removed value still remains present in the old
    context.

    Args:
        key: The key of the entry to remove
        context: The context to copy, if None, the current context is used
    """
    if context is None:
        context = get_current()
    new_values = context.copy()
    new_values.pop(key, None)
    return Context(new_values)


def get_current() -> Context:
    """To access the context associated with program execution,
    the RuntimeContext API provides a function which takes no arguments
    and returns a RuntimeContext.
    """

    global _RUNTIME_CONTEXT  # pylint: disable=global-statement
    if _RUNTIME_CONTEXT is None:
        # FIXME use a better implementation of a configuration manager to avoid having
        # to get configuration values straight from environment variables
        if version_info < (3, 5):
            # contextvars are not supported in 3.4, use thread-local storage
            default_context = "threadlocal_context"
        else:
            default_context = "contextvars_context"

        configured_context = environ.get(
            "OPENTELEMETRY_CONTEXT", default_context
        )  # type: str
        try:
            _RUNTIME_CONTEXT = next(
                iter_entry_points("opentelemetry_context", configured_context)
            ).load()()
        except Exception:  # pylint: disable=broad-except
            logger.error("Failed to load context: %s", configured_context)

    return _RUNTIME_CONTEXT.get_current()  # type:ignore


def set_current(context: Context) -> Context:
    """To associate a context with program execution, the Context
    API provides a function which takes a Context.

    Args:
        context: The context to use as current.
    """
    old_context = get_current()
    _RUNTIME_CONTEXT.set_current(context)  # type:ignore
    return old_context


def with_current_context(
    func: typing.Callable[..., "object"]
) -> typing.Callable[..., "object"]:
    """Capture the current context and apply it to the provided func."""

    caller_context = get_current()

    def call_with_current_context(
        *args: "object", **kwargs: "object"
    ) -> "object":
        try:
            backup = get_current()
            set_current(caller_context)
            return func(*args, **kwargs)
        finally:
            set_current(backup)

    return call_with_current_context



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import opentelemetry.context.propagation.httptextformat as httptextformat
import opentelemetry.trace as trace
from opentelemetry.context.propagation.tracecontexthttptextformat import (
    TraceContextHTTPTextFormat,
)

_T = typing.TypeVar("_T")


def extract(
    get_from_carrier: httptextformat.Getter[_T], carrier: _T
) -> trace.SpanContext:
    """Load the parent SpanContext from values in the carrier.

    Using the specified HTTPTextFormatter, the propagator will
    extract a SpanContext from the carrier. If one is found,
    it will be set as the parent context of the current span.

    Args:
        get_from_carrier: a function that can retrieve zero
            or more values from the carrier. In the case that
            the value does not exist, return an empty list.
        carrier: and object which contains values that are
            used to construct a SpanContext. This object
            must be paired with an appropriate get_from_carrier
            which understands how to extract a value from it.
    """
    return get_global_httptextformat().extract(get_from_carrier, carrier)


def inject(
    tracer: trace.Tracer,
    set_in_carrier: httptextformat.Setter[_T],
    carrier: _T,
) -> None:
    """Inject values from the current context into the carrier.

    inject enables the propagation of values into HTTP clients or
    other objects which perform an HTTP request. Implementations
    should use the set_in_carrier method to set values on the
    carrier.

    Args:
        set_in_carrier: A setter function that can set values
            on the carrier.
        carrier: An object that contains a representation of HTTP
            headers. Should be paired with set_in_carrier, which
            should know how to set header values on the carrier.
    """
    get_global_httptextformat().inject(
        tracer.get_current_span(), set_in_carrier, carrier
    )


_HTTP_TEXT_FORMAT = (
    TraceContextHTTPTextFormat()
)  # type: httptextformat.HTTPTextFormat


def get_global_httptextformat() -> httptextformat.HTTPTextFormat:
    return _HTTP_TEXT_FORMAT


def set_global_httptextformat(
    http_text_format: httptextformat.HTTPTextFormat,
) -> None:
    global _HTTP_TEXT_FORMAT  # pylint:disable=global-statement
    _HTTP_TEXT_FORMAT = http_text_format



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The OpenTelemetry metrics API describes the classes used to report raw
measurements, as well as metrics with known aggregation and labels.

The `Meter` class is used to construct `Metric` s to record raw statistics
as well as metrics with predefined aggregation.

See the `metrics api`_ spec for terminology and context clarification.

.. _metrics api:
    https://github.com/open-telemetry/opentelemetry-specification/blob/master/specification/api-metrics.md


"""
import abc
from typing import Callable, Dict, Optional, Sequence, Tuple, Type, TypeVar

from opentelemetry.util import loader

ValueT = TypeVar("ValueT", int, float)


class DefaultMetricHandle:
    """The default MetricHandle.

    Used when no MetricHandle implementation is available.
    """

    def add(self, value: ValueT) -> None:
        """No-op implementation of `CounterHandle` add.

        Args:
            value: The value to add to the handle.
        """

    def set(self, value: ValueT) -> None:
        """No-op implementation of `GaugeHandle` set.

        Args:
            value: The value to set to the handle.
        """

    def record(self, value: ValueT) -> None:
        """No-op implementation of `MeasureHandle` record.

        Args:
            value: The value to record to the handle.
        """


class CounterHandle:
    def add(self, value: ValueT) -> None:
        """Increases the value of the handle by ``value``.

        Args:
            value: The value to add to the handle.
        """


class GaugeHandle:
    def set(self, value: ValueT) -> None:
        """Sets the current value of the handle to ``value``.

        Args:
            value: The value to set to the handle.
        """


class MeasureHandle:
    def record(self, value: ValueT) -> None:
        """Records the given ``value`` to this handle.

        Args:
            value: The value to record to the handle.
        """


class LabelSet(abc.ABC):
    """A canonicalized set of labels useful for preaggregation

    Re-usable LabelSet objects provide a potential optimization for scenarios
    where handles might not be effective. For example, if the LabelSet will be
    re-used but only used once per metrics, handles do not offer any
    optimization. It may best to pre-compute a canonicalized LabelSet once and
    re-use it with the direct calling convention. LabelSets are immutable and
    should be opaque in implementation.
    """


class DefaultLabelSet(LabelSet):
    """The default LabelSet.

    Used when no LabelSet implementation is available.
    """


class Metric(abc.ABC):
    """Base class for various types of metrics.

    Metric class that inherit from this class are specialized with the type of
    handle that the metric holds.
    """

    @abc.abstractmethod
    def get_handle(self, label_set: LabelSet) -> "object":
        """Gets a handle, used for repeated-use of metrics instruments.

        Handles are useful to reduce the cost of repeatedly recording a metric
        with a pre-defined set of label values. All metric kinds (counter,
        gauge, measure) support declaring a set of required label keys. The
        values corresponding to these keys should be specified in every handle.
        "Unspecified" label values, in cases where a handle is requested but
        a value was not provided are permitted.

        Args:
            label_set: `LabelSet` to associate with the returned handle.
        """


class DefaultMetric(Metric):
    """The default Metric used when no Metric implementation is available."""

    def get_handle(self, label_set: LabelSet) -> "DefaultMetricHandle":
        """Gets a `DefaultMetricHandle`.

        Args:
            label_set: `LabelSet` to associate with the returned handle.
        """
        return DefaultMetricHandle()

    def add(self, value: ValueT, label_set: LabelSet) -> None:
        """No-op implementation of `Counter` add.

        Args:
            value: The value to add to the counter metric.
            label_set: `LabelSet` to associate with the returned handle.
        """

    def set(self, value: ValueT, label_set: LabelSet) -> None:
        """No-op implementation of `Gauge` set.

        Args:
            value: The value to set the gauge metric to.
            label_set: `LabelSet` to associate with the returned handle.
        """

    def record(self, value: ValueT, label_set: LabelSet) -> None:
        """No-op implementation of `Measure` record.

        Args:
            value: The value to record to this measure metric.
            label_set: `LabelSet` to associate with the returned handle.
        """


class Counter(Metric):
    """A counter type metric that expresses the computation of a sum."""

    def get_handle(self, label_set: LabelSet) -> "CounterHandle":
        """Gets a `CounterHandle`."""
        return CounterHandle()

    def add(self, value: ValueT, label_set: LabelSet) -> None:
        """Increases the value of the counter by ``value``.

        Args:
            value: The value to add to the counter metric.
            label_set: `LabelSet` to associate with the returned handle.
        """


class Gauge(Metric):
    """A gauge type metric that expresses a pre-calculated value.

    Gauge metrics have a value that is either ``Set`` by explicit
    instrumentation or observed through a callback. This kind of metric
    should be used when the metric cannot be expressed as a sum or because
    the measurement interval is arbitrary.
    """

    def get_handle(self, label_set: LabelSet) -> "GaugeHandle":
        """Gets a `GaugeHandle`."""
        return GaugeHandle()

    def set(self, value: ValueT, label_set: LabelSet) -> None:
        """Sets the value of the gauge to ``value``.

        Args:
            value: The value to set the gauge metric to.
            label_set: `LabelSet` to associate with the returned handle.
        """


class Measure(Metric):
    """A measure type metric that represent raw stats that are recorded.

    Measure metrics represent raw statistics that are recorded.
    """

    def get_handle(self, label_set: LabelSet) -> "MeasureHandle":
        """Gets a `MeasureHandle` with a float value."""
        return MeasureHandle()

    def record(self, value: ValueT, label_set: LabelSet) -> None:
        """Records the ``value`` to the measure.

        Args:
            value: The value to record to this measure metric.
            label_set: `LabelSet` to associate with the returned handle.
        """


MetricT = TypeVar("MetricT", Counter, Gauge, Measure)


# pylint: disable=unused-argument
class Meter(abc.ABC):
    """An interface to allow the recording of metrics.

    `Metric` s are used for recording pre-defined aggregation (gauge and
    counter), or raw values (measure) in which the aggregation and labels
    for the exported metric are deferred.
    """

    @abc.abstractmethod
    def record_batch(
        self,
        label_set: LabelSet,
        record_tuples: Sequence[Tuple["Metric", ValueT]],
    ) -> None:
        """Atomically records a batch of `Metric` and value pairs.

        Allows the functionality of acting upon multiple metrics with
        a single API call. Implementations should find metric and handles that
        match the key-value pairs in the label tuples.

        Args:
            label_set: The `LabelSet` associated with all measurements in
                the batch. A measurement is a tuple, representing the `Metric`
                being recorded and the corresponding value to record.
            record_tuples: A sequence of pairs of `Metric` s and the
                corresponding value to record for that metric.
        """

    @abc.abstractmethod
    def create_metric(
        self,
        name: str,
        description: str,
        unit: str,
        value_type: Type[ValueT],
        metric_type: Type[MetricT],
        label_keys: Sequence[str] = (),
        enabled: bool = True,
    ) -> "Metric":
        """Creates a ``metric_kind`` metric with type ``value_type``.

        Args:
            name: The name of the metric.
            description: Human-readable description of the metric.
            unit: Unit of the metric values.
            value_type: The type of values being recorded by the metric.
            metric_type: The type of metric being created.
            label_keys: The keys for the labels with dynamic values.
            enabled: Whether to report the metric by default.
        Returns: A new ``metric_type`` metric with values of ``value_type``.
        """

    @abc.abstractmethod
    def get_label_set(self, labels: Dict[str, str]) -> "LabelSet":
        """Gets a `LabelSet` with the given labels.

        Args:
            labels: A dictionary representing label key to label value pairs.

        Returns: A `LabelSet` object canonicalized using the given input.
        """


class DefaultMeter(Meter):
    """The default Meter used when no Meter implementation is available."""

    def record_batch(
        self,
        label_set: LabelSet,
        record_tuples: Sequence[Tuple["Metric", ValueT]],
    ) -> None:
        pass

    def create_metric(
        self,
        name: str,
        description: str,
        unit: str,
        value_type: Type[ValueT],
        metric_type: Type[MetricT],
        label_keys: Sequence[str] = (),
        enabled: bool = True,
    ) -> "Metric":
        # pylint: disable=no-self-use
        return DefaultMetric()

    def get_label_set(self, labels: Dict[str, str]) -> "LabelSet":
        # pylint: disable=no-self-use
        return DefaultLabelSet()


# Once https://github.com/python/mypy/issues/7092 is resolved,
# the following type definition should be replaced with
# from opentelemetry.util.loader import ImplementationFactory
ImplementationFactory = Callable[[Type[Meter]], Optional[Meter]]

_METER = None
_METER_FACTORY = None


def meter() -> Meter:
    """Gets the current global :class:`~.Meter` object.

    If there isn't one set yet, a default will be loaded.
    """
    global _METER, _METER_FACTORY  # pylint:disable=global-statement

    if _METER is None:
        # pylint:disable=protected-access
        try:
            _METER = loader._load_impl(Meter, _METER_FACTORY)  # type: ignore
        except TypeError:
            # if we raised an exception trying to instantiate an
            # abstract class, default to no-op tracer impl
            _METER = DefaultMeter()
        del _METER_FACTORY

    return _METER


def set_preferred_meter_implementation(factory: ImplementationFactory) -> None:
    """Set the factory to be used to create the meter.

    See :mod:`opentelemetry.util.loader` for details.

    This function may not be called after a meter is already loaded.

    Args:
        factory: Callback that should create a new :class:`Meter` instance.
    """
    global _METER, _METER_FACTORY  # pylint:disable=global-statement

    if _METER:
        raise RuntimeError("Meter already loaded.")

    _METER_FACTORY = factory



# Copyright 2019, OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import string
import typing
from contextlib import contextmanager

from opentelemetry.context import get_value, set_current, set_value
from opentelemetry.context.context import Context

PRINTABLE = frozenset(
    itertools.chain(
        string.ascii_letters, string.digits, string.punctuation, " "
    )
)


class EntryMetadata:
    """A class representing metadata of a DistributedContext entry

    Args:
        entry_ttl: The time to live (in service hops) of an entry. Must be
                   initially set to either :attr:`EntryMetadata.NO_PROPAGATION`
                   or :attr:`EntryMetadata.UNLIMITED_PROPAGATION`.
    """

    NO_PROPAGATION = 0
    UNLIMITED_PROPAGATION = -1

    def __init__(self, entry_ttl: int) -> None:
        self.entry_ttl = entry_ttl


class EntryKey(str):
    """A class representing a key for a DistributedContext entry"""

    def __new__(cls, value: str) -> "EntryKey":
        return cls.create(value)

    @staticmethod
    def create(value: str) -> "EntryKey":
        # pylint: disable=len-as-condition
        if not 0 < len(value) <= 255 or any(c not in PRINTABLE for c in value):
            raise ValueError("Invalid EntryKey", value)

        return typing.cast(EntryKey, value)


class EntryValue(str):
    """A class representing the value of a DistributedContext entry"""

    def __new__(cls, value: str) -> "EntryValue":
        return cls.create(value)

    @staticmethod
    def create(value: str) -> "EntryValue":
        if any(c not in PRINTABLE for c in value):
            raise ValueError("Invalid EntryValue", value)

        return typing.cast(EntryValue, value)


class Entry:
    def __init__(
        self, metadata: EntryMetadata, key: EntryKey, value: EntryValue
    ) -> None:
        self.metadata = metadata
        self.key = key
        self.value = value


class DistributedContext:
    """A container for distributed context entries"""

    def __init__(self, entries: typing.Iterable[Entry]) -> None:
        self._container = {entry.key: entry for entry in entries}

    def get_entries(self) -> typing.Iterable[Entry]:
        """Returns an immutable iterator to entries."""
        return self._container.values()

    def get_entry_value(self, key: EntryKey) -> typing.Optional[EntryValue]:
        """Returns the entry associated with a key or None

        Args:
            key: the key with which to perform a lookup
        """
        if key in self._container:
            return self._container[key].value
        return None


class DistributedContextManager:
    def get_current_context(
        self, context: typing.Optional[Context] = None
    ) -> typing.Optional[DistributedContext]:
        """Gets the current DistributedContext.

        Returns:
            A DistributedContext instance representing the current context.
        """

    @contextmanager  # type: ignore
    def use_context(
        self, context: DistributedContext
    ) -> typing.Iterator[DistributedContext]:
        """Context manager for controlling a DistributedContext lifetime.

        Set the context as the active DistributedContext.

        On exiting, the context manager will restore the parent
        DistributedContext.

        Args:
            context: A DistributedContext instance to make current.
        """
        # pylint: disable=no-self-use
        yield context


_DISTRIBUTED_CONTEXT_KEY = "DistributedContext"


def distributed_context_from_context(
    context: typing.Optional[Context] = None,
) -> DistributedContext:
    return get_value(_DISTRIBUTED_CONTEXT_KEY, context)  # type: ignore


def with_distributed_context(
    dctx: DistributedContext, context: typing.Optional[Context] = None
) -> None:
    set_current(set_value(_DISTRIBUTED_CONTEXT_KEY, dctx, context=context))
